'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _typeof2 = require('babel-runtime/helpers/typeof');

var _typeof3 = _interopRequireDefault(_typeof2);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 *  @fileOverview   Class Fluxter.
 *  @description    Simple container of unidirectional data flow.
 *  @author         Vasily Shilov <shcoder.ru@ya.ru> (https://github.com/uxter)
 *  @licence        ISC (https://github.com/uxter/fluxter/blob/master/LICENSE.txt)
 *  {@link          https://github.com/uxter/fluxter GitHub}
 *  {@link          https://www.npmjs.com/package/fluxter NPM}
 */

/**
 * Class Fluxter
 * @class Fluxter
 * @public
 *
 * @example
 * let store = new Fluxter({
 *     user: {
 *         logged: false,
 *         name: null,
 *         token: null
 *     }
 * });
 */
var Fluxter = function () {

    /**
     * @constructor
     * @param {Object=} state - Initial state
     */
    function Fluxter() {
        var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
        (0, _classCallCheck3.default)(this, Fluxter);

        checkArgumentType(state, 'first', 'object');
        this.reducers = [];
        this.middlewares = [];
        this.actions = {};
        this.handlers = [];
        this.state = state;
    }

    /**
     * Add reducer to store
     * @method addReducer
     * @param {String} stateKey - A state field
     * @param {Function} reducer - A pure function
     *
     * @example
     * store.addReducer('user', (state = {}, actionName, actionData) => {
     *     if (actionName === 'login') {
     *         return {
     *             ...state,
     *             logged: actionData.logged,
     *             name: actionData.name,
     *             token: actionData.token
     *         };
     *     }
     *     return state;
     * });
     */


    (0, _createClass3.default)(Fluxter, [{
        key: 'addReducer',
        value: function addReducer(stateKey, reducer) {
            checkArgumentType(stateKey, 'first', 'string');
            checkArgumentType(reducer, 'second', 'function');
            this.reducers.push({
                stateKey: stateKey,
                reducer: reducer
            });
        }

        /**
         * Add middleware to store
         * @method addMiddleware
         * @param {Function} middleware - A middleware function
          * @example
         * store.addMiddleware('user', (store, actionName, actionData, next) => {
         *     if (actionName === 'login') {
         *         SomeRepository.check(actionData).then(data => next({
         *             ...data,
         *             ...actionData
         *         }));
         *     } else {
         *         next(actionData);
         *     }
         * });
         */

    }, {
        key: 'addMiddleware',
        value: function addMiddleware(middleware) {
            checkArgumentType(middleware, 'first', 'function');
            this.middlewares.push(middleware);
        }

        /**
         * Add action to store
         * @method addAction
         * @param {String} actionName - An action name
         * @param {Function} action - An action creator function
         *
         * @example
         * store.addAction('login', (name, password) => ({
         *     name,
         *     password
         * }));
         */

    }, {
        key: 'addAction',
        value: function addAction(actionName, action) {
            checkArgumentType(actionName, 'first', 'string');
            checkArgumentType(action, 'second', 'function');
            this.actions[actionName] = action;
        }

        /**
         * Dispatch action to store
         * @method dispatch
         * @param {String} actionName - An action name
         * @param {...*} args - Arguments that take action
         *
         * @example
         * store.dispatch('login', 'example-name', 'example-password');
         */

    }, {
        key: 'dispatch',
        value: function dispatch(actionName) {
            var _actions,
                _this = this;

            if (typeof this.actions[actionName] !== 'function') {
                throw new Error('Action "' + actionName + '" was not specified.');
            }

            for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
                args[_key - 1] = arguments[_key];
            }

            var actionData = (_actions = this.actions)[actionName].apply(_actions, args);
            callMiddlewares.call(this, actionName, actionData, function (actionData) {
                callReducers.call(_this, actionName, actionData);
                fire.call(_this);
            });
        }

        /**
         * Add handler
         * @method subscribe
         * @param {Function} handler - Change state handler
         *
         * @example
         * store.subscribe(store => view.setState(store.state));
         */

    }, {
        key: 'subscribe',
        value: function subscribe(handler) {
            checkArgumentType(handler, 'first', 'function');
            this.handlers.push(handler);
        }
    }]);
    return Fluxter;
}();

/**
 * Call middlewares
 * @function callMiddlewares
 * @param {String} actionName - An action name
 * @param {*} actionData - An action data
 * @param {Function} done - Done handler
 * @private
 */


exports.default = Fluxter;
function callMiddlewares(actionName, actionData, done) {
    var store = this;
    var index = -1;
    /**
     * Call next middleware or done
     * @function next
     * @param {*} actionData - An action data
     */
    function next(actionData) {
        index++;
        if (index === store.middlewares.length) {
            done(actionData);
        } else {
            store.middlewares[index](store, actionName, actionData, next);
        }
    }
    next(actionData);
}

/**
 * Call reducers
 * @function callReducers
 * @param {String} actionName - An action name
 * @param {*} actionData - An action data
 * @private
 */
function callReducers(actionName, actionData) {
    var _this2 = this;

    this.reducers.forEach(function (item) {
        _this2.state[item.stateKey] = item.reducer(_this2.state[item.stateKey], actionName, actionData);
    });
}

/**
 * Call handlers
 * @function fire
 * @private
 */
function fire() {
    var _this3 = this;

    this.handlers.forEach(function (handler) {
        return handler(_this3);
    });
}

/**
 * Check type of argument
 * @function checkArgumentType
 * @param {*} argument
 * @param {String} position - first, second, ...
 * @param {String} type - string, function, ...
 * @throws {TypeError}
 * @private
 */
function checkArgumentType(argument, position, type) {
    var wrongTypeOf = (typeof argument === 'undefined' ? 'undefined' : (0, _typeof3.default)(argument)) !== type;
    var isCheckObject = type === 'object';
    var n = isCheckObject ? 'n' : '';
    if (isCheckObject && (wrongTypeOf || argument.constructor !== Object) || wrongTypeOf) {
        throw new TypeError('A ' + position + ' argument must be a' + n + ' ' + type + '.');
    }
}